@@ -7,6 +7,7 @@ our $PERL_VERSION = $^V;
 $PERL_VERSION =~ s|^v||;
 
 use OpenGL qw/ :all /;
+use OpenGL::Config;     # for build information
 
 eval 'use OpenGL::Image 1.03';  # Need to use OpenGL::Image 1.03 or higher!
 my $hasImage = !$@;
@@ -71,7 +72,7 @@ use constant DO_TESTS => 0;
 
 # Run in Game Mode
 my $gameMode;
-if ($#ARGV and lc($ARGV[0]) eq 'gamemode')
+if (scalar(@ARGV) and lc($ARGV[0]) eq 'gamemode')
 {
   $gameMode = $ARGV[1] || '';
 }
@@ -90,8 +91,8 @@ my $hasFBO = 0;
 my $hasVBO = 0;
 my $hasFragProg = 0;
 my $hasImagePointer = 0;
-my $idleTime = time();
-my $idleSecsMax = 30;
+my $idleTime = $hasHires ? gettimeofday() : time();
+my $idleSecsMax = 5;
 my $er;
 
 # Window and texture IDs, window width and height.
@@ -337,10 +338,8 @@ sub ourPrintString
 # Does everything needed before losing control to the main
 # OpenGL event loop.
 
-sub ourInit
+sub ourInitVertexBuffers
 {
-  my ($Width, $Height) = @_;
-
   # Set initial colors for rainbow face
   for (my $i=0; $i<16; $i++)
   {
@@ -351,6 +350,8 @@ sub ourInit
   # Initialize VBOs if supported
   if ($hasVBO)
   {
+    printf("Using VBOs\n");
+
     ($VertexObjID,$NormalObjID,$ColorObjID,$TexCoordObjID,$IndexObjID) =
       glGenBuffersARB_p(5);
 
@@ -398,12 +399,21 @@ sub ourInit
   }
   else
   {
+    print "Using classic Vertex Buffers\n";
     glVertexPointer_p(3, $verts);
     glNormalPointer_p($norms);
     $colors->assign($rainbow_offset,@rainbow);
     glColorPointer_p(4, $colors);
     glTexCoordPointer_p(2, $texcoords);
   }
+  print "-- done\n";
+}
+
+sub ourInit
+{
+  my ($Width, $Height) = @_;
+
+  printf("\nUsing POGL v$OpenGL::VERSION\n");
 
   # Build texture.
   ($TextureID_image,$TextureID_FBO) = glGenTextures_p(2);
@@ -413,6 +423,9 @@ sub ourInit
   # Initialize shaders.
   ourInitShaders();
 
+  # Initialize vertex buffers
+  ourInitVertexBuffers();
+
   # Initialize rendering parameters
   glEnable(GL_TEXTURE_2D);
   glDisable(GL_LIGHTING);
@@ -456,12 +469,12 @@ sub ourBuildTextures
 
   # Build Image Texture
   ($TextureID_image,$TextureID_FBO) = glGenTextures_p(2);
-  print "\n";
 
   # Use OpenGL::Image to load texture
   if ($hasImage && -e $Tex_File)
   {
     my $img = new OpenGL::Image(source=>$Tex_File);
+    die $@ if $@;
     my($eng,$ver) = $img->Get('engine','version');
     print "Using OpenGL::Image - $eng v$ver\n";
 
@@ -611,6 +624,8 @@ sub ourBuildTextures
   # Build FBO texture
   if ($hasFBO)
   {
+    printf("Using FBOs\n");
+
     ($FrameBufferID) = glGenFramebuffersEXT_p(1);
     ($RenderBufferID) = glGenRenderbuffersEXT_p(1);
 
@@ -627,7 +642,7 @@ sub ourBuildTextures
     glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
       GL_TEXTURE_2D, $TextureID_FBO, 0);
     glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, $RenderBufferID);
-    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24,
+    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24_ARB,
       $Tex_Width, $Tex_Height);
     glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
       GL_RENDERBUFFER_EXT, $RenderBufferID);
@@ -655,9 +670,22 @@ sub ourInitShaders
 {
   # Setup Vertex/Fragment Programs to render FBO texture
 
-  # Use OpenGL::Shader
-  if ($hasShader && ($Shader = new OpenGL::Shader()))
+  if ($hasShader)
   {
+    my $version = $OpenGL::Shader::VERSION;
+    printf("Using OpenGL::Shader v$version\n");
+    my $types = OpenGL::Shader->GetTypes();
+    my @types = keys(%$types);
+    printf("This installation supports the following shader types: %s\n", join(',', @types));
+
+    # Use OpenGL::Shader
+    $Shader = new OpenGL::Shader();
+    if (!$Shader)
+    {
+      printf("Unable to instantiate OpenGL::Shader\n");
+      return;
+    }
+
     my $type = $Shader->GetType();
     my $ext = lc($type);
 
@@ -673,10 +701,10 @@ sub ourInitShaders
       print "$stat\n";
     }
   }
-
   # Fall back to doing it manually
-  if ($hasFragProg)
+  elsif ($hasFragProg)
   {
+    printf("Using native OpenGL ARB Shader functions\n");
     ($VertexProgID,$FragProgID) = glGenProgramsARB_p(2);
 
     # NOP Vertex shader
@@ -763,11 +791,13 @@ sub ourInitShaders
 sub cbRenderScene
 {
   # Quit if inactive
-  if ($idleSecsMax < (time()-$idleTime))
+  my $time = $hasHires ? gettimeofday() : time();
+  my $time_to_exit = $idleSecsMax - ($time-$idleTime);
+  if ($time_to_exit <= 0 )
   {
     print "Idle timeout; completing test\n";
     ourCleanup();
-    exit(0);
+    return quit("render callback");
   }
 
   my $buf; # For our strings.
@@ -961,6 +991,10 @@ sub cbRenderScene
   # But, for fun, let's make the text partially transparent too.
   glColor4f(0.6,1.0,0.6,.75);
 
+  $buf = sprintf "TIME TO EXIT: %.1fs", $time_to_exit;
+  my $bufwidth = 6 * length $buf;
+  glRasterPos2i($Window_Width-4-$bufwidth,2); ourPrintString(GLUT_BITMAP_HELVETICA_12,$buf);
+
   # Render our various display mode settings.
   $buf = sprintf "Mode: %s", $TexModesStr[$Curr_TexMode];
   glRasterPos2i(2,2); ourPrintString(GLUT_BITMAP_HELVETICA_12,$buf);
@@ -1082,6 +1116,7 @@ sub Inset
   {
     my $frame = new OpenGL::Image(engine=>'Magick',
       width=>$Inset_Width, height=>$Inset_Height);
+    die $@ if $@;
     my($fmt,$size) = $frame->Get('gl_format','gl_type');
 
     glReadPixels_c( $Capture_X, $Capture_Y, $Inset_Width, $Inset_Height,
@@ -1134,6 +1169,7 @@ sub Save
 # Cleanup routine
 sub ourCleanup
 {
+  print "Starting cleanup ...\n";
   # Disable app
   glutHideWindow();
   glutKeyboardUpFunc();
@@ -1148,13 +1184,16 @@ sub ourCleanup
   # Now you can destroy window
   if (defined($gameMode))
   {
+    print "Leaving game mode.\n";
     glutLeaveGameMode();
   }
   else
   {
+    print "Destroying window.\n";
     glutDestroyWindow($Window_ID);
   }
   undef($Window_ID);
+  print "Cleanup completed.\n";
 }
 
 sub ReleaseResources
@@ -1209,7 +1248,7 @@ sub cbKeyPressed
   if ($key == 27 or $c eq 'Q')
   {
     ourCleanup();
-    exit(0);
+    return quit("key press callback");
   }
   elsif ($c eq 'B')
   {
@@ -1288,7 +1327,7 @@ sub cbKeyPressed
     printf "KP: No action for %d.\n", $key;
   }
 
-  $idleTime = time();
+  $idleTime = $hasHires ? gettimeofday() : time();
 }
 
 # ------
@@ -1327,7 +1366,7 @@ sub cbSpecialKeyPressed
     printf "SKP: No action for %d.\n", $key;
   }
 
-  $idleTime = time();
+  $idleTime = $hasHires ? gettimeofday() : time();
 }
 
 # ------
@@ -1433,7 +1472,7 @@ sub cbMouseClick
     print "\n";
   }
 
-  $idleTime = time();
+  $idleTime = $hasHires ? gettimeofday() : time();
 }
 
 sub GetKeyModifier
@@ -1464,7 +1503,7 @@ sub cbResizeScene
   $Window_Width  = $Width;
   $Window_Height = $Height;
 
-  $idleTime = time();
+  $idleTime = $hasHires ? gettimeofday() : time();
 }
 
 sub cbWindowStat
@@ -1480,6 +1519,31 @@ sub cbClose
   ReleaseResources();
 }
 
+# this is a little complicated
+# Using freeglut, doing a straight exit crashes on some systems, mostly observed
+# on windows, likely due to thread issues.
+# However on non-freeglut systems the proper way of using glutLeaveMainLoop is
+# not available.
+# So the proper one needs to be chosen.
+# However while exit exits the thread, glutLeaveMainLoop only sets a flag for
+# the event loop, thus we must take care to return when using it. Additionally
+# any use of quit() ALSO needs to return.
+#
+#   return quit();
+#
+sub quit {
+  my ($context) = @_;
+  $context ||= "<unknown context>";
+  print "Exiting in $context using ";
+  if (OpenGL::_have_freeglut()) {
+    print "glutLeaveMainLoop (freeglut)\n";
+    glutLeaveMainLoop();
+    return;
+  }
+  print "perl exit(0)\n";
+  exit(0);
+}
+
 # ------
 # The main() function.  Inits OpenGL.  Calls our own init function,
 # then passes control onto OpenGL.
@@ -1490,7 +1554,7 @@ glutInit();
 # To see OpenGL drawing, take out the GLUT_DOUBLE request.
 glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
 
-if ($^O ne 'MSWin32') { # skip these MODE checks on win32, they don't work
+if ($^O ne 'MSWin32' and $OpenGL::Config->{DEFINE} !~ /-DHAVE_W32API/) { # skip these MODE checks on win32, they don't work
 
    if (not glutGet(GLUT_DISPLAY_MODE_POSSIBLE))
    {
@@ -1608,11 +1672,14 @@ Press 'c' to capture/save a RGBA targa file.
 # Pass off control to OpenGL.
 # Above functions are called as appropriate.
 if (OpenGL::_have_freeglut()) {
+   print "Setting window close to trigger return from mainloop (freeglut).\n";
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE,GLUT_ACTION_GLUTMAINLOOP_RETURNS)
 }
 
+print "Entering glutMainLoop\n";
 glutMainLoop();
+print "Returned from glutMainLoop\n";
 
-print "FreeGLUT returned from MainLoop\n";
+print "Exiting in main thread\n";
 
 __END__
